@@ -1,7 +1,8 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * PhoneManager Listener
  * Copyright (C) 2003-2004 Edd Dumbill <edd@usefulinc.com>
- * Copyright (C) 2005 Bastien Nocera <hadess@hadess.net>
+ * Copyright (C) 2005-2007 Bastien Nocera <hadess@hadess.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,6 +18,7 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
  */
+
 #include <config.h>
 
 #include <stdlib.h>
@@ -78,9 +80,12 @@ struct _PhonemgrListener
 
 	guint connected : 1;
 	guint terminated : 1;
+	guint debug : 1;
 
 	/* Whether the driver supports GN_OP_OnSMS */
 	guint supports_sms_notif : 1;
+	/* Whether or not the driver supports GN_OP_GetPowersource */
+	guint supports_power_source : 1;
 };
 
 static void phonemgr_listener_class_init (PhonemgrListenerClass *klass);
@@ -99,10 +104,53 @@ enum {
 	LAST_SIGNAL
 };
 
+enum {
+	PROP_0,
+	PROP_DEBUG,
+};
+
 static int phonemgr_listener_signals[LAST_SIGNAL] = { 0 } ;
 
 G_DEFINE_TYPE (PhonemgrListener, phonemgr_listener, G_TYPE_OBJECT)
 
+static void
+phonemgr_listener_get_property (GObject *object,
+				 guint property_id,
+				 GValue *value,
+				 GParamSpec *pspec)
+{
+	PhonemgrListener *l;
+
+	l = PHONEMGR_LISTENER (object);
+
+	switch (property_id) {
+	case PROP_DEBUG:
+		g_value_set_boolean (value, l->debug != FALSE);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+	}
+}
+
+static void
+phonemgr_listener_set_property (GObject *object,
+				guint property_id,
+				const GValue *value,
+				GParamSpec *pspec)
+{
+	PhonemgrListener *l;
+
+	l = PHONEMGR_LISTENER (object);
+
+	switch (property_id) {
+	case PROP_DEBUG:
+		l->debug = g_value_get_boolean (value);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+	}
+}
+
 static void
 phonemgr_listener_class_init (PhonemgrListenerClass *klass)
 {
@@ -111,6 +159,8 @@ phonemgr_listener_class_init (PhonemgrListenerClass *klass)
 	parent_class = g_type_class_ref (G_TYPE_OBJECT);
 
 	object_class = (GObjectClass*) klass;
+	object_class->set_property = phonemgr_listener_set_property;
+	object_class->get_property = phonemgr_listener_get_property;
 
 	phonemgr_listener_signals[MESSAGE_SIGNAL] =
 		g_signal_new ("message",
@@ -156,6 +206,11 @@ phonemgr_listener_class_init (PhonemgrListenerClass *klass)
 			      2,
 			      G_TYPE_INT, G_TYPE_BOOLEAN);
 
+	g_object_class_install_property (object_class,
+					 PROP_DEBUG,
+					 g_param_spec_boolean ("debug", NULL, NULL,
+							       FALSE, G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+
 	object_class->finalize = phonemgr_listener_finalize;
 
 	klass->message = NULL;
@@ -208,19 +263,33 @@ phonemgr_listener_emit_message (PhonemgrListener *l, gn_sms *message)
 
 	text = NULL;
 
-	//FIXME use the encoding as coming out of the txt message
+	/* The data should be in whatever the locale's encoding is,
+	 * if it doesn't fit the default GSM alphabet (roughly ISO8859-1)
+	 * as a design decision in gnokii */
 	origtext = (char *) message->user_data[0].u.text;
-	if (g_utf8_validate (origtext, -1, NULL) == FALSE) {
+
+	if (gn_char_def_alphabet (origtext)) {
 		GError *err = NULL;
 
-		text = g_convert (origtext,
-				strlen (origtext),
-				"utf-8", "iso-8859-1",
-				NULL, NULL, &err);
+		text = g_convert (origtext, -1,
+				  "UTF-8", "ISO8859-1",
+				  NULL, NULL, &err);
+
 		if (err != NULL) {
-			g_warning ("Conversion error: %d %s",
-					err->code, err->message);
+			g_warning ("Conversion error from GSM default alphabet: %d %s", err->code, err->message);
 			g_error_free (err);
+			g_free (text);
+			text = g_strdup (origtext);
+		}
+	} else if (g_utf8_validate (origtext, -1, NULL) == FALSE) {
+		GError *err = NULL;
+
+		text = g_locale_to_utf8 (origtext, -1, NULL, NULL, &err);
+
+		if (err != NULL) {
+			g_warning ("Conversion error: %d %s", err->code, err->message);
+			g_error_free (err);
+			g_free (text);
 			text = g_strdup (origtext);
 		}
 	}
@@ -240,9 +309,9 @@ phonemgr_listener_emit_message (PhonemgrListener *l, gn_sms *message)
 }
 
 PhonemgrListener *
-phonemgr_listener_new (void)
+phonemgr_listener_new (gboolean debug)
 {
-	return PHONEMGR_LISTENER (g_object_new (PHONEMGR_TYPE_LISTENER, NULL));
+	return PHONEMGR_LISTENER (g_object_new (PHONEMGR_TYPE_LISTENER, "debug", debug, NULL));
 }
 
 #ifndef DUMMY
@@ -253,6 +322,7 @@ phonemgr_listener_init (PhonemgrListener *l)
 	l->mutex = g_mutex_new ();
 	l->driver = NULL;
 	l->batterylevel = 1;
+	l->supports_power_source = TRUE;
 	l->powersource = GN_PS_BATTERY;
 }
 
@@ -294,7 +364,7 @@ phonemgr_listener_connect (PhonemgrListener *l, char *device, GError **error)
 		}
 	}
 
-	l->phone_state = phonemgr_utils_connect (device, NULL, channel, error);
+	l->phone_state = phonemgr_utils_connect (device, NULL, channel, l->debug, error);
 	if (l->phone_state == NULL) {
 		//FIXME
 		return FALSE;
@@ -310,7 +380,7 @@ phonemgr_listener_connect (PhonemgrListener *l, char *device, GError **error)
 	if (strcmp (l->driver, PHONEMGR_DEFAULT_DRIVER) != 0) {
 		phonemgr_utils_disconnect (l->phone_state);
 		phonemgr_utils_free (l->phone_state);
-		l->phone_state = phonemgr_utils_connect (device, l->driver, channel, error);
+		l->phone_state = phonemgr_utils_connect (device, l->driver, channel, l->debug, error);
 		if (l->phone_state == NULL) {
 			//FIXME
 			return FALSE;
@@ -488,14 +558,20 @@ phonemgr_listener_battery_poll (PhonemgrListener *l)
 	(&l->phone_state->data)->power_source = &powersource;
 	(&l->phone_state->data)->battery_unit = &battery_unit;
 
-	if (gn_sm_functions(GN_OP_GetPowersource, &l->phone_state->data, &l->phone_state->state) != GN_ERR_NONE)
+	/* Some drivers will use the same function for battery level and power source, so optimise.
+	 * Make sure to get the battery level first, as most drivers implement only this one */
+	if (gn_sm_functions(GN_OP_GetBatteryLevel, &l->phone_state->data, &l->phone_state->state) != GN_ERR_NONE)
+		return;
+
+	if (powersource == -1 && l->supports_power_source != FALSE) {
+		if (gn_sm_functions(GN_OP_GetPowersource, &l->phone_state->data, &l->phone_state->state) != GN_ERR_NONE) {
+			g_message ("driver or phone doesn't support getting the power source");
+			l->supports_power_source = FALSE;
+			powersource = GN_PS_BATTERY;
+		}
+	} else if (l->supports_power_source == FALSE) {
 		powersource = GN_PS_BATTERY;
-
-	/* Some drivers will use the same function for battery level and power source, so optimise */
-	if (batterylevel == -1)
-		if (gn_sm_functions(GN_OP_GetBatteryLevel, &l->phone_state->data, &l->phone_state->state) != GN_ERR_NONE)
-			return;
-
+	}
 
 	if (batterylevel != l->batterylevel || powersource != l->powersource) {
 		AsyncSignal *signal;
@@ -529,9 +605,9 @@ phonemgr_listener_set_sms_notification (PhonemgrListener *l, gboolean state)
 		l->phone_state->data.callback_data = l;
 		if (gn_sm_functions (GN_OP_OnSMS, &l->phone_state->data, &l->phone_state->state) == GN_ERR_NONE) {
 			l->supports_sms_notif = TRUE;
-			g_message ("driver supports sms notifications");
+			g_message ("driver and phone support sms notifications");
 		} else {
-			g_message ("driver doesn't support sms nots");
+			g_message ("driver or phone doesn't support sms notifications");
 		}
 	} else {
 		if (l->supports_sms_notif == FALSE)
@@ -629,7 +705,7 @@ void
 phonemgr_listener_queue_message (PhonemgrListener *l,
 		const char *number, const char *message)
 {
-	char *mstr;
+	char *mstr, *iso;
 	GError *err = NULL;
 	gn_sms sms;
 	gn_error error;
@@ -638,26 +714,42 @@ phonemgr_listener_queue_message (PhonemgrListener *l,
 	g_return_if_fail (number != NULL);
 	g_return_if_fail (message != NULL);
 
-	/* Lock the phone */
+	/* Lock the phone and set up for SMS sending */
 	g_mutex_lock (l->mutex);
-
 	gn_data_clear(&l->phone_state->data);
+	gn_sms_default_submit(&sms);
 
-	mstr = g_convert (message, strlen (message),
-			"iso-8859-1", "utf-8",
-			NULL, NULL, &err);
+	/* Default GSM alphabet is a subset of ISO8859-1,
+	 * so try that */
+	iso = g_convert (message, -1,
+			 "ISO8859-1", "UTF-8",
+			 NULL, NULL, &err);
+	if (err != NULL) {
+		g_clear_error (&err);
+		g_free (iso);
+		iso = NULL;
+	}
 
-	g_message ("mstr '%s'", mstr);
+	/* If the message contains characters not in the
+	 * default GSM alaphabet, we pass it as UTF-8 encoding instead */
+	if (iso && gn_char_def_alphabet((unsigned char *) iso)) {
+		mstr = iso;
+		sms.dcs.u.general.alphabet = GN_SMS_DCS_DefaultAlphabet;
+	} else {
+		g_free (iso);
+		/* gnokii will convert to UCS-2 itself, but we pass UTF-8 */
+		mstr = g_strdup (message);
+		sms.dcs.u.general.alphabet = GN_SMS_DCS_UCS2;
+	}
 
 	if (err != NULL) {
 		g_warning ("Conversion error: %d %s", err->code, err->message);
+		g_free (mstr);
 		g_error_free (err);
 		g_mutex_unlock (l->mutex);
 		return;
 	}
 
-	gn_sms_default_submit(&sms);
-
 	/* Set the destination number */
 	g_strlcpy (sms.remote.number, number, sizeof(sms.remote.number));
 	if (sms.remote.number[0] == '+')
@@ -678,18 +770,17 @@ phonemgr_listener_queue_message (PhonemgrListener *l,
 	sms.user_data[0].type = GN_SMS_DATA_Text;
 	g_strlcpy ((char *) sms.user_data[0].u.text, mstr, strlen (mstr) + 1);
 	sms.user_data[0].length = strlen (mstr);
-
-	/* Set the message encoding */
-	if (!gn_char_def_alphabet(sms.user_data[0].u.text))
-		sms.dcs.u.general.alphabet = GN_SMS_DCS_UCS2;
-	sms.user_data[1].type = GN_SMS_DATA_None;
 	g_free (mstr);
+	sms.user_data[1].type = GN_SMS_DATA_None;
 
 	l->phone_state->data.sms = &sms;
 
 	/* Actually send the message */
 	error = gn_sms_send (&l->phone_state->data, &l->phone_state->state);
 
+	/* Remove the reference to SMS */
+	l->phone_state->data.sms = NULL;
+
 	/* Unlock the phone */
 	g_mutex_unlock (l->mutex);
 }
